
package nl.madline.opendp500;

import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.net.Socket;
import java.util.Arrays;
import java.util.Hashtable;

/**
 * @author svmaris
 */
public class JDPClientHandler implements Runnable {
    static final int STREAM_BUFFER_SIZE = 32768;
    static final int COMMAND_BUFFER_SIZE = 256;

    private Socket client;
    private JDPServer dserver;
    private PrintWriter out;
    private InputStreamReader in;
	private String ipOfClient;
	
    /**
     * Default constructor.
     * 
	 * @param dserver The server that created the connection
	 * @param client The connected player
	 */
	public JDPClientHandler(JDPServer dserver, Socket client) {
        this.dserver = dserver;
        this.client = client;
        this.in = null;
        this.out = null;

		ipOfClient = client.getInetAddress().getHostAddress();

        try {
            this.in = new InputStreamReader(this.client.getInputStream());
        } catch (IOException e) {
            System.err.println("Failed to create inputstream");
            e.printStackTrace();
        }

        try {
            this.out = new PrintWriter(this.client.getOutputStream(),true);
        } catch (IOException e) {
            System.err.println("Failed to create outputstream");
            e.printStackTrace();
        }

    }

    /* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
        String socketInfo;
        socketInfo = this.client.getRemoteSocketAddress().toString();

        this.dserver.debugMsg("New client connection from " + socketInfo);

        try {
            // Keep a network buffer of 255 chars,
            // this is maybe too low for very deep-nested files
            // or insanely long filenames. Will adjust later when
            // needed.

            char [] buf = new char[JDPClientHandler.COMMAND_BUFFER_SIZE];
            while(this.in.read(buf) > 0) {
                parseCommand(String.valueOf(buf).trim());

                // Clear input buffer
                buf = new char[JDPClientHandler.COMMAND_BUFFER_SIZE];
            }
        } catch (IOException e) {
            dserver.deletePlayer(ipOfClient);
        }
    }

    /**
     * Process a command that was received from the player
     * 
	 * @param input The String received from the player
	 */
	private void parseCommand(String input) {
    	dserver.setLastCommand(input);
        dserver.debugMsg(input);
        if(input.startsWith("LIST")) {
            dserver.debugMsg("Directory listing requested");
            
            String [] parts = input.split(" ");
            String [] parts2 = input.split("\\|");

            dserver.setCurrentMediaType(parts[1]);

            if(parts2.length == 1) {
                displayMainCommands();
            }
            else {
                String command = parts2[1];
                String [] commandParts = command.split("/");
                String mainCommand = commandParts[0];
                mainCommand = this.dserver.getMainCommand(mainCommand);

                if(mainCommand != null) {
                    if(mainCommand.equals("PLAYLISTS")) {
                        if(commandParts.length == 1) {
                            displayPlaylistCommands();
                        }
                        else {
                            String subCommand = dserver.getPlaylistCommand(commandParts[1]);
                            if(subCommand.equals("VIEWPLAYLIST")) {
                                if(commandParts.length == 3) {
                                    displayPlaylist(commandParts[2],false);
                                }
                                else {
                                    displayPlaylists();
                                }
                            }
                            else if(subCommand.equals("NEWPLAYLIST")) {
                                JDPPlaylist list = dserver.addPlaylist();
                                doOutput("Playlist " + list.getName() + " created","",1);
                            }
                            else if(subCommand.equals("DELETEPLAYLIST")) {
                                if(commandParts.length == 3) {
                                    dserver.removePlaylist(commandParts[2]);
                                    doOutput("Playlist " + commandParts[2] + " deleted","",1);
                                }
                                else {
                                    displayPlaylists();
                                }
                            }
                            else if(subCommand.equals("DELETEFROMPLAYLIST")) {
                                if(commandParts.length == 2) {
                                    displayPlaylists();
                                }
                                else if(commandParts.length == 3) {
                                    displayPlaylist(commandParts[2],true);
                                }
                                else if(commandParts.length > 3) {
                                    JDPPlaylist list = dserver.getPlaylist(commandParts[2]);
                                    String filename = commandParts[3];
                                    int number = new Integer(filename.substring(0,filename.indexOf("."))).intValue();
                                    list.removeFile(number);
                                    doOutput("File " + filename + " removed from playlist " + list.getName(),"",1);
                                }
                            }
                            else if(subCommand.equals("ADDPLAYINGFILE")) {
                                if(commandParts.length == 3) {
                                    JDPPlaylist list = dserver.getPlaylist(commandParts[2]);
                                    if(dserver.getNowPlaying() == null) {
                                        doOutput("There is no file playing now","",1);
                                    }
                                    else {
                                        list.addFile(dserver.getNowPlaying());
                                        doOutput("Current playing file added to " + commandParts[2],"",1);
                                    }
                                }
                                else {
                                    displayPlaylists();
                                }
                            }
                            else if(subCommand.equals("ADDTOPLAYLIST")) {
                                if(commandParts.length == 2) {
                                    displayPlaylists();
                                }
                                else if(commandParts.length > 3) {
                                    String path = "";

                                    for (int i = 3; i < commandParts.length; i++) {
                                        String commandPart = commandParts[i];
                                        path += "/" + commandPart;
                                    }

                                    //path = path.substring(0,path.length()-1);
                                    String pathPrefix = dserver.getPath(dserver.getCurrentMediaType());

                                    dserver.debugMsg("Browsing " + path);
                                    if(new File(pathPrefix + path).isDirectory()) {
                                        displayLibrary(dserver.getCurrentMediaType(),path,true);
                                    }
                                    else {
                                        dserver.getPlaylist(commandParts[2]).addFile(pathPrefix + path);
                                        doOutput("File added to playlist " + commandParts[2],"",1);
                                    }
                                }
                                else {
                                    displayLibrary(dserver.getCurrentMediaType(),"",true);
                                }
                            }
                            else {
                                displayNotImplemented();
                            }
                        }
                    }
                    else if(mainCommand.equals("OPTIONS")) {
                        if(commandParts.length == 1) {
                            displayOptionCommands();
                        }
                        else {
                            String subCommand = dserver.getOptionCommand(commandParts[1]);
                            doOutput(subCommand,subCommand,1);
                        }
                    }
                    else if(mainCommand.equals("ABOUT")) {
                        doOutput("--------------------------------------------------","",2);
                        doOutput(dserver.getRoot(),"",2);
                        doOutput("--------------------------------------------------","",2);
                    }
                    else if(mainCommand.equals("LIBRARY")) {
                        command = command.replaceFirst(commandParts[0],"");
                        displayLibrary(parts[1], command, false);
                    }
                    else if(mainCommand.equals("RECENT")) {
                        displayPlaylist("Most Recently Played", false);
                    }
                    else {
                        displayNotImplemented();
                    }
                }
                else {
                    doOutput("Unknown Command - Send bugreport to svmaris@e-village.nl","",1);
                }
            }
            closeConnection();
        }

        if(input.startsWith("SIZE")) {
            String filename = input.substring(5,input.length()-1);
            String filesize = new Long(new File(filename).length()).toString();

            this.dserver.debugMsg("Requesting filesize, returning " + filesize);

            // Pad up to 15 bytes, because the player wants that
            while(filesize.length() < 15) {
                filesize = "0" + filesize;
            }

            this.out.print(filesize);
            this.out.flush();

        }

        if(input.startsWith("ACTION")) {
            // I'm not really sure why this command is used.
            // It is called 2 times: 'ACTION 2' and 'ACTION 1'
            // and responding with '200', like the PC-Link Software
            // does, seems to do the trick
            this.out.write("200");
            this.out.flush();
        }

        if(input.startsWith("GET")) {
            // Looks odd, but it works for all my files ;)
            String [] parts = input.split("\\|");
            boolean disconnectAfterPlay = parts.length == 3 && parts[2].length() == 0 ? true : false;
            String filename = parts[0].substring(4);

            String [] parts2 = parts[1].trim().split(" ");
            long offset = new Long(parts2[0]).longValue();
            long length = new Long(parts2[1]).longValue();

            try {
                RandomAccessFile raf = new RandomAccessFile(filename,"r");

                if(offset > 0) {
                    raf.seek(offset);
                }

                if(length == 0) {
                    length = raf.length()-offset;
                }

                this.dserver.debugMsg("Getting filename " + filename + ", offset " + offset + ", length " + length);
                dserver.setNowPlaying(filename);

                DataOutputStream dos = new DataOutputStream(client.getOutputStream());

                byte [] b = null;
                // Keep a default buffer of 32768 bytes to keep
                // the disk/cpu load low. If the dp500 wants a smaller
                // piece (to check codec and id3v2-tag), adjust the
                // buffer
                if(length < (long)JDPClientHandler.STREAM_BUFFER_SIZE) {
                    b = new byte[(int)length];
                }
                else {
                    b = new byte[JDPClientHandler.STREAM_BUFFER_SIZE];
                }

                int rest = (int)(length % b.length);
                int bytesRead;

                for(long l=0;l<length/b.length;l++) {
                    bytesRead = raf.read(b);
                    dos.write(b,0,bytesRead);
                    dos.flush();
                }

                if(rest > 0) {
                    b = new byte[rest];
                    bytesRead = raf.read(b);
                    dos.write(b,0,bytesRead);
                    dos.flush();
                }

				raf.close();
				
                if(disconnectAfterPlay) {
                    this.dserver.debugMsg("Disconnect after play");
                    this.dserver.debugMsg("Adding to recently played files");
                    dserver.getPlaylist("Most Recently Played").addFile(filename);
                    this.dserver.debugMsg("Adding to popular music");
                    dserver.getPlaylist("Most Played").addFile(filename);
                    closeConnection();
                }

            } catch (Exception e) {
                this.dserver.debugMsg("Stopped playing");
                closeConnection();
            }
        }
    }

    /**
     * Display the contents of a playlist on the DP-500.
     * 
	 * @param list The list to be displayed
	 * @param playlistMode If true, the files will be displayed as directories, so
	 * this function can be used as a file selection tool to add/remove files from
	 * a given playlist. If false, the contents of the playlist will be presented as
	 * files, so the DP-500 will try to play them when selected.
	 */
	private void displayPlaylist(JDPPlaylist list, boolean playlistMode) {
        String [] contents = list.getFiles();

        if(contents.length == 0) {
            doOutput("Playlist is empty","",1);
        }

        for (int i = 0; i < contents.length; i++) {
            String content = null;
            content = contents[i];

            String filename = content.substring(content.lastIndexOf('/')+1,content.length());
            if(playlistMode) {
                doOutput((i+1)+". "+filename,content,1);
            }
            else {
                doOutput(filename,content,0);
            }
        }
    }

	/**
	 * Display the contents of a playlist on the DP-500 by its name.
	 * 
	 * @param list The list to be displayed
	 * @param playlistMode If true, the files will be displayed as directories, so
	 * this function can be used as a file selection tool to add/remove files from
	 * a given playlist. If false, the contents of the playlist will be presented as
	 * files, so the DP-500 will try to play them when selected.
	 */
	private void displayPlaylist(String name, boolean playlistMode) {
        JDPPlaylist list = dserver.getPlaylist(name);
        displayPlaylist(list,playlistMode);
    }

    /**
	 * Get all playlists from the server and display them on the DP-500
	 */
	private void displayPlaylists() {
        String [] playlists = dserver.getPlaylists(dserver.getCurrentMediaType());

        if(playlists.length == 0) {
            doOutput("No playlists found","",1);
        }
        else {
            for (int i = 0; i < playlists.length; i++) {
                String playlist = playlists[i];

                doOutput(playlist,playlist,1);
            }
        }
    }

    /**
     * Display the contents of the library in a given subdirectory
     * 
	 * @param mediaType The media type to be displayed (AUDIO/VIDEO/PICTURE)
	 * @param subDirectory The subdictory to be displayed
	 * @param playlistMode If true, the files will be displayed as directories, so
	 * this function can be used as a file selection tool to add/remove files from
	 * a given playlist. If false, the contents of the playlist will be presented as
	 * files, so the DP-500 will try to play them when selected.
	 */
	private void displayLibrary(String mediaType, String subDirectory, boolean playlistMode) {
        String rootPath = dserver.getPath(mediaType);
        if(rootPath.endsWith("/")) {
            rootPath = rootPath.substring(0,rootPath.length()-1);
        }

        String fullPath = rootPath + subDirectory;

        this.dserver.debugMsg("Requesting directory listing on " + fullPath);

        File f = new File(fullPath);

        if(f.exists()) {
            File [] contents = f.listFiles();
            Arrays.sort(contents);
            for (int i = 0; i < contents.length; i++) {
                File content = contents[i];
                if(!content.getName().startsWith(".")) {
                    if(content.isDirectory() || playlistMode == true) {
                        doOutput(content.getName(),fullPath + "/" + content.getName(),1);
                    }
                    else {
                        doOutput(content.getName(),fullPath + "/" + content.getName(),0);
                    }
                }
            }
        }
        else {
            doOutput("File not found","",1);
        }
    }

    /**
     * Display the contents of a given Hashtable, used to display the main-menu,
     * the playlist-menu and the options
     * 
	 * @param h The hashtable to be displayed
	 */
	private void displayHashtable(Hashtable h) {
        Object [] commandsArray = h.keySet().toArray();

        Arrays.sort(commandsArray);

        for (int i = 0; i < commandsArray.length; i++) {
            String s = (String)commandsArray[i];

            doOutput(s,s,1);
        }
    }

    /**
	 * Display the commands for the playlist on the DP-500
	 */
	private void displayPlaylistCommands() {
        Hashtable commands = dserver.getPlaylistCommands();
        displayHashtable(commands);
    }

    /**
	 * Display the main menu on the DP-500
	 */
	private void displayMainCommands() {
        Hashtable commands = dserver.getMainCommands();
        displayHashtable(commands);
    }

    /**
	 * Display the option menu on the DP-500
	 */
	private void displayOptionCommands() {
        Hashtable commands = dserver.getOptionCommands();
        displayHashtable(commands);
    }

    /**
	 * Stub to display when a function is not implemented yet
	 */
	private void displayNotImplemented() {
        doOutput("Not Implemented Yet","",1);
    }

    /**
     * Write a line of text to the DP-500.
     * 
	 * @param display The text that has to be displayed on the screen
	 * @param command The current directory
	 * @param type The type of the current line, which is '1' for 
	 * a directory and '2' for a file.
	 */
	private void doOutput(String display, String command, int type) {
        this.out.println(display + "|" + command + "|" + type + "|");
    }

    /**
	 * Close the connection to the player.
	 */
	private void closeConnection() {
        this.dserver.debugMsg("Closing connection");
        try {
			dserver.deletePlayer(ipOfClient);

            this.in.close();
            this.out.close();
            this.client.close();            
        } catch(Exception e) {
            this.dserver.debugMsg("Couldn't close connection");
        }
    }
}
