/*
 * Created on Apr 22, 2003
 *
 */
package nl.madline.opendp500;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Properties;

/**
 * This is a representation of the configuration file. Methods
 * are created to get/set options and save the file to disk
 * 
 * @author svmaris
 */
public class JDPConfig {
	private Properties config;
	private File configFile;
	/**
	 * Default contructor
	 * 
	 * @param f The configuration file as a File-object
	 */
	public JDPConfig(File f) {
		configFile = f;
		config = new Properties();
		
		try {
			BufferedReader data = new BufferedReader(new FileReader(f));
			
			String line;
			int lineNumber = 0;
			while((line = data.readLine()) != null) {
				lineNumber++;

				// Discard line if it starts with [ or #
				if(line.trim().length() > 0 &&
				 	!line.startsWith("[") &&
				    !line.startsWith("#")) {
						
					String [] parts = line.split("=");
					
					if(parts.length == 2) {
						setProperty(parts[0].trim(),parts[1].trim());				
					}
					else {
						System.out.println("Parse error in config on line " + lineNumber);
						System.exit(1);
					}
				}
			}
		
			
		} catch (FileNotFoundException e) {
			System.out.println("Error: can not find the configuration file");
			System.exit(1);
		} catch (IOException e) {
			System.out.println("Error while reading the configuration file");
			System.exit(1);
		}
	}

	/**
	 * Set a property in the configuration to 'value'. Please note that
	 * the changes are not written to disk yet. You will have to call
	 * save() to do that.
	 * 
	 * @param name The variable name in the configuration file
	 * @param value The new value of the setting
	 */
	public void setProperty(String name, String value) {
		config.setProperty(name,value);		
	}


	/**
	 * Return the current configuration as a Properties object
	 * 
	 * @return properties All configuration options and their values as
	 * a Properties object
	 */
	public Properties getProperties() {
		return config;
	}
	
	/**
	 * Return the value of a given property in the configuration
	 * 
	 * @param property The name of the property
	 * @return property The value of the property
	 */
	public String getProperty(String property) {
		return config.getProperty(property);
	}
	
	
	/**
	 * Save the current configuration from memory to disk,
	 * using the filename this object was created with. I
	 * built it so that the formatting of the original
	 * configuration file will stay intact (comments, order of
	 * configuration items, etc). Only the changed properties
	 * will be replaced.
	 */
	public void save() {
		try {
			Properties oldProperties = new JDPConfig(configFile).getProperties();
			BufferedReader data = new BufferedReader(new FileReader(configFile));
			
			String configData = "";
			String line = null;
			
			while((line = data.readLine()) != null) {
				configData += line + "\n";
			}
			
			Enumeration e = oldProperties.keys();
			String currentLine = null;
				
			while(e.hasMoreElements()) {
				currentLine = (String)e.nextElement();
				
				if(!oldProperties.getProperty(currentLine).equals(config.getProperty(currentLine)) ) {
					configData = replace(configData,currentLine+"="+oldProperties.getProperty(currentLine),currentLine+"="+config.getProperty(currentLine));
				}
			}
				
			PrintWriter p = new PrintWriter(new FileWriter(configFile));
			p.write(configData);
			p.close();
		} catch (FileNotFoundException e) {
			System.out.println("The configuration file can not be found.");
			System.exit(1);
		} catch (IOException e) {
			System.out.println("An error occured while writing the configuration file");
			System.exit(1);
		}
	}
	
	/**
	 * Replaces a substring with another
	 * 
	 * @param str Source string
	 * @param pattern What to replace
	 * @param replace Replace with...
	 * @return result The resulting string
	 */
	private String replace(String str, String pattern, String replace) {
		int s = 0;
		int e = 0;
		StringBuffer result = new StringBuffer();
    
		while ((e = str.indexOf(pattern, s)) >= 0) {
			result.append(str.substring(s, e));
			result.append(replace);
			s = e+pattern.length();
		}

		result.append(str.substring(s));
		return result.toString();
	}

	/**
	 * @return file The configuration-file as a File-object
	 */
	public File getConfigFile() {
		return configFile;
	}
}
