package nl.madline.opendp500;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import nl.madline.opendp500.web.JDPWebServer;

/**
 * @author svmaris
 */
public class JDPServer {
    private static int PORT = 8000;

    private ServerSocket server;
    private String nowPlaying, currentMediaType;
    private String root;
    private String lastCommand;
    private Hashtable mainCommands = new Hashtable();
    private Hashtable playlistCommands = new Hashtable();
    private Hashtable optionCommands = new Hashtable();
	private JDPConfig dpconfig;
    private Vector playlists = new Vector();
	private Vector connectedPlayers = new Vector();
	
    /**
     * Default constructor
     * 
	 * @param audioPath Path to mp3-files
	 * @param videoPath Path to Video-files
	 * @param picturePath Path to Pictures
	 */
	public JDPServer(JDPConfig config) {    	
		dpconfig = config;

		this.debugMsg("Audio path   : " + config.getProperty("path.audio"));
		this.debugMsg("Video path   : " + config.getProperty("path.video"));
		this.debugMsg("Picture path : " + config.getProperty("path.picture"));
		    	    	    	
    	// Launch webserver
    	if(config.getProperty("webserver.enabled").equals("true")) {
	    	new Thread(new JDPWebServer(this)).start();
    	}
    	
        this.server = null;

        addPlaylist("Most Recently Played");
        addPlaylist("Most Played");

        this.root = JDP500.APPNAME + " " + JDP500.VERSION + " by " + JDP500.AUTHOR;

        this.mainCommands.put("1 - Library","LIBRARY");
        this.mainCommands.put("2 - Playlists","PLAYLISTS");
        this.mainCommands.put("3 - Options","OPTIONS");
        this.mainCommands.put("4 - About","ABOUT");

        this.playlistCommands.put("2.1 - View Playlists","VIEWPLAYLIST");
        this.playlistCommands.put("2.2 - New Playlist","NEWPLAYLIST");
        this.playlistCommands.put("2.3 - Add File to Playlist","ADDTOPLAYLIST");
        this.playlistCommands.put("2.4 - Delete File from Playlist","DELETEFROMPLAYLIST");
        this.playlistCommands.put("2.5 - Delete Playlist","DELETEPLAYLIST");
        this.playlistCommands.put("2.6 - Rename Playlist","RENAMEPLAYLIST");
        this.playlistCommands.put("2.7 - Add Currently Playing File to Playlist","ADDPLAYINGFILE");

        this.optionCommands.put("5.1 - Remove all playlists","CLEARPLAYLISTS");
        this.optionCommands.put("5.2 - Clear Most Recently Played","CLEARRECENT");
        this.optionCommands.put("5.3 - Clear Most Played","CLEARPOPULAR");

        try {
            this.server = new ServerSocket(JDPServer.PORT);
        }
        catch(IOException e) {
            System.err.println("Failed to open ServerSocket on port " + JDPServer.PORT);
            System.exit(1);
        }

        this.debugMsg("DPServer is running on port " + JDPServer.PORT);

        while(true) {
            try {
                Socket client = this.server.accept();
                String ip = client.getInetAddress().getHostAddress();
                
                if(isAllowedToConnect(ip)) {
	                JDPClientHandler handler;
                
                
    	            handler = new JDPClientHandler(this,client);

					addPlayer(client.getInetAddress().getHostAddress());
				
            	    // Make it multi-threaded, so we can browse and play
                	// files at the same time.
                	new Thread(handler).start();
                }
                else {
                	this.debugMsg("Client " + ip + " is not allowed to connect.");
                	client.close();
                	client = null;
                }
            }
            catch(IOException e) {
                System.err.println("Failed to accept client");
            }
        }
    }

    /**
     * Add a playlist with a default name and return
     * the playlist
     * 
	 * @return playlist The newly created playlist
	 */
	public JDPPlaylist addPlaylist() {
        return addPlaylist("Unnamed " + (playlists.size()+1));
    }

    /**
     * Remove playlist by name
     * 
	 * @param name The name of the playlist to be removed
	 */
	public void removePlaylist(String name) {
        for (int i = 0; i < playlists.size(); i++) {
            JDPPlaylist playlist = (JDPPlaylist) playlists.elementAt(i);

            if(playlist.getName().equals(name)) {
                playlists.remove(playlist);
                break;
            }
        }
    }

    /**
     * Adds a playlist with the given name and returns it.
     * This is not used on the player yet, because there's
     * currently no way to enter the name on the DP-500
     * 
	 * @param name The name for the playlist
	 * @return playlist The newly created playlist
	 */
	public JDPPlaylist addPlaylist(String name) {
        JDPPlaylist list = new JDPPlaylist(getCurrentMediaType(),name);
        playlists.add(list);
        return list;
    }

    /**
     * Get all playlists from a certain media-type
     * 
	 * @param mediaType The mediatype (AUDIO, VIDEO, PICTURE)
	 * @return playlists An array of Strings containing all the playlists
	 */
	public String [] getPlaylists(String mediaType) {
        Vector v = new Vector();

        for (int i = 0; i < playlists.size(); i++) {
            JDPPlaylist playlist = (JDPPlaylist) playlists.elementAt(i);

            if(playlist.getMediaType() == null) {
                playlist.setMediaType(this.currentMediaType);
            }

            if(playlist.getMediaType().equals(mediaType)) {
                v.add(playlist);
            }
        }

        String [] returnValue = new String[v.size()];

        for (int i = 0; i < v.size(); i++) {
            JDPPlaylist jdpPlaylist = (JDPPlaylist) playlists.elementAt(i);

            returnValue[i] = jdpPlaylist.getName();
        }

        return returnValue;
    }

    /**
     * Get the playlist with a given name
     * 
	 * @param name The name of the playlist
	 * @return playlist The playlist with that name
	 */
	public JDPPlaylist getPlaylist(String name) {
        for (int i = 0; i < playlists.size(); i++) {
            JDPPlaylist jdpPlaylist = (JDPPlaylist) playlists.elementAt(i);

            if(jdpPlaylist.getName().equals(name)) {
                return jdpPlaylist;
            }
        }

        return null;
    }

    /**
     * Get the default audiopath as given on the commandline
     * 
	 * @return audioPath The base path to the audio-files
	 */
	public String getAudioPath() {
        return dpconfig.getProperty("path.audio");
    }

    /**
     * Set the audioPath
     * 
	 * @param audioPath The base path to the audio-files
	 */
	public void setAudioPath(String audioPath) {
		dpconfig.setProperty("path.audio",audioPath);
    }

    /**
     * Get the default videopath as given on the commandline
     * 
	 * @return videoPath The base path to the video-files
	 */
	public String getVideoPath() {
        return dpconfig.getProperty("path.video");
    }

    /**
     * Set the videoPath
     * 
	 * @param videoPath The base path to the video-files
	 */
	public void setVideoPath(String videoPath) {
        dpconfig.setProperty("path.video",videoPath);
    }

    /**
     * Get the default picturepath as given on the commandline
     * 
	 * @return picturePath The base path to the pictures
	 */
	public String getPicturePath() {
        return dpconfig.getProperty("path.picture");
    }

    /**
     * Set the picturePath
     * 
	 * @param picturePath The base path to the pictures
	 */
	public void setPicturePath(String picturePath) {
        dpconfig.setProperty("path.picture",picturePath);
    }

    /**
     * Get APPNAME and VERSION-string for the about-screen
     * 
	 * @return root Application name and version information
	 */
	public String getRoot() {
        return this.root;
    }

    /**
     * Returns the internal mainCommand.
     * mainCommands are the functions that are reached
     * from the main menu on the DP-500
     * 
	 * @param command The command received from the player
	 * @return command The translated (internal) command
	 */
	public String getMainCommand(String command) {
        this.debugMsg("getMainCommand(" + command +")");
        if(this.mainCommands.containsKey(command)) {
            return (String)this.mainCommands.get(command);
        }
        else {
            return null;
        }
    }

    /**
     * Returns the internal playlistCommand.
     * playlistCommands are the functions that are reached
     * from the Playlist menu on the DP-500
     * 
	 * @param command The command received from the player
	 * @return command The translated (internal) command
	 */
	public String getPlaylistCommand(String command) {
        this.debugMsg("getPlaylistCommand(" + command +")");
        if(this.playlistCommands.containsKey(command)) {
            return (String)this.playlistCommands.get(command);
        }
        else {
            return null;
        }
    }

    /**
     * Returns the internal optionCommand.
     * optionCommands are the functions that are reached
     * from the Option menu on the DP-500
     * 
	 * @param command The command received from the player
	 * @return command The translated (internal) command
	 */
	public String getOptionCommand(String command) {
        this.debugMsg("getOptionCommand(" + command +")");
        if(this.optionCommands.containsKey(command)) {
            return (String)this.optionCommands.get(command);
        }
        else {
            return null;
        }
    }

    /**
     * Get a Hashtable with the translation-table for the
     * player mainCommands and internal commands
     * 
	 * @return commands The hashtable containing the translation
	 */
	public Hashtable getMainCommands() {
        return this.mainCommands;
    }

    /**
     * Get a Hashtable with the translation-table for the
     * player playlistCommands and internal commands
     * 
	 * @return commands The hashtable containing the translation
	 */
	public Hashtable getPlaylistCommands() {
        return playlistCommands;
    }

    /**
     * Get a HAshtable with the translation-table for the
     * player optionCommands and internal commands
     * 
	 * @return commands The hashtable containing the translation
	 */
	public Hashtable getOptionCommands() {
        return optionCommands;
    }

    /**
     * Get the root path of a certain mediaType
     * 
	 * @param mediaType The media type (AUDIO/VIDEO/PICTURE)
	 * @return path The root path of the given media type
	 */
	public String getPath(String mediaType) {
        if(mediaType.equals("AUDIO")) {
            return getAudioPath();
        }
        else if(mediaType.equals("VIDEO")) {
            return getVideoPath();
        }
        else if(mediaType.equals("PICTURE")) {
            return getPicturePath();
        }
        else {
            return null;
        }
    }

    /**
     * Returns the currently playing file, or 'Not Playling'
     * if there's not a file playing at the moment
     * 
	 * @return nowPlaying The currently playing file
	 */
	public String getNowPlaying() {
    	if(nowPlaying == null) {
    		return "Not playing";
    	}
    	else {
    	    return nowPlaying;
    	}
    }

    /**
     * Set the nowPlaying-property
     * 
	 * @param nowPlaying The name of the file that is currently playing
	 */
	public void setNowPlaying(String nowPlaying) {
        this.debugMsg("Setting nowPlaying to " + nowPlaying);
        this.nowPlaying = nowPlaying;
    }

    /**
     * Get the current media type
     * 
	 * @return mediaType The current media type of the player
	 */
	public String getCurrentMediaType() {
        return currentMediaType;
    }

    /**
     * Set the current media type
     * 
	 * @param currentMediaType The media type the player is in right now
	 */
	public void setCurrentMediaType(String currentMediaType) {
        this.currentMediaType = currentMediaType;
    }
    
    /**
     * Get a String with all the connected players
     * 
	 * @return connectedPlayers The IP-addresses of the connected players, seperated by a space
	 */
	public String getConnectedPlayers() {
    	String returnValue = "";
    	
    	if(connectedPlayers.size() == 0) {
    		returnValue = "No connected players found";
    	}
    	else {    	
    		Enumeration e = connectedPlayers.elements();
    		
    		while(e.hasMoreElements()) {
    			returnValue += (String)e.nextElement() + " ";
    		}
    	}
    	
    	return returnValue;
    }
    
    /**
     * Add a player to the connected players-list
     * 
	 * @param ip The IP-address of the new player
	 */
	public void addPlayer(String ip) {
    	connectedPlayers.add(ip);
    }
    
    /**
     * Remove a connected player by IP-address
     * 
	 * @param ip The IP-address of the player to be removed
	 */
	public void deletePlayer(String ip) { 
		if(connectedPlayers.indexOf(ip) > -1) {
			connectedPlayers.remove(ip);			   	
		}
    }
    
	/**
	 * Return the last command received by the player
	 * 
	 * @return lastCommand The last command received by the player
	 */
	public String getLastCommand() {
		if(lastCommand == null) {
			return "Nothing received yet";
		}
		else {		
			return lastCommand;
		}
	}

	/**
	 * Set the lastCommand to a given value.
	 * 
	 * @param string The command received from the player	
	 */
	public void setLastCommand(String string) {
		lastCommand = string;
	}

	/**
	 * @param property The property to be fetched
	 * @return propery The requested property
	 */
	public String getProperty(String property) {
		return dpconfig.getProperty(property);
	}
	
	/**
	 * @param ip The IP of the connecting client
	 * @return allowed true if allowed to connect
	 */
	public boolean isAllowedToConnect(String ip) {
		String allow = dpconfig.getProperty("ip.allow");
		
		if(allow == null || allow.equals("")) {
			return true;
		}
		else {
			String [] ips = allow.split(";");
			
			for (int i = 0; i < ips.length; i++) {
				if(ip.startsWith(ips[i])) {
					return true;
				}
			}
		}
		
		return false;
	}
	
	/**
	 * Tell the config to save itself to disk
	 */
	public void saveConfig() {
		dpconfig.save();
	}
	
	public JDPConfig getConfig() {
		return dpconfig;		
	}

	/**
	 * Reload the configuration from the file
	 */
	public void reloadConfig() {
		dpconfig = new JDPConfig(dpconfig.getConfigFile());		
	}
	
	/**
	 * Print a debugmessage to the console if debugging is enabled
	 */
	public void debugMsg(String message) {
		if(this.getProperty("debug.enabled").equals("true")) {
			System.out.println(message);
		}	
	}
}
